/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

namespace MpcSharp
{

    using System;
    using System.Collections;
    using Gtk;
    using MPD;

    /*
     * A list of all the available tabs.
     */
    enum TabType
    {
        Playlist,
        Library,
        Find,
        Cover,
        Playlists
    }


    class MpcSharpApp
    {
        Glade.XML gxml;
        [Glade.Widget] Window window;
        [Glade.Widget] Statusbar statusbar;
        [Glade.Widget] Notebook notebook;
        [Glade.Widget] Expander expander;
        [Glade.Widget] Label expander_label;
        [Glade.Widget] EventBox expander_eventbox;
        [Glade.Widget] Viewport expander_vp;
        [Glade.Widget] HBox hbox_row1;
        [Glade.Widget] HBox hbox_row2;

        double PanBeginX = 0;
        double PanBeginHadj = 0;

        /*
         * The formatter is used to parse the StatusFormat syntax.
         */
        Format formatter;
        Format Formatter
        {
            get
            {
                if (formatter == null)
                    formatter = new Format(Settings.StatusFormat);
                return formatter;
            }
        }


        /*
         * The main entry point is just a simple wrapper that starts the main
         * Gtk event loop.
         */
        public static void Main(string[] args)
        {
            Gtk.Application.Init();
            new MpcSharpApp(args);
            Gtk.Application.Run();
        }


        public MpcSharpApp(string[] args)
        {
            gxml = Glade.XML.FromAssembly("gui.glade", "window", null);
            gxml.Autoconnect(this);

            // Add all small GuiWidgets
            SetGuiWidgets(hbox_row1, Settings.GuiRow1);
            SetGuiWidgets(hbox_row2, Settings.GuiRow2);

            // Restore old window size and position
            window.Move(Settings.WinPosX, Settings.WinPosY);
            window.Resize(Settings.WinWidth, Settings.WinHeight);

            // Read config and show all chosen tabs
            foreach (string tab in Settings.Tabs)
            {
                //Type type;
                try
                {
                    TabType type = (TabType) Enum.Parse(typeof(TabType), tab);
                    //type = Type.GetType("MpcSharp." + tab);
                    AddTab(type);
                }
                catch (ArgumentException e)
                {
                    Console.WriteLine(e);
                }
            }
            notebook.Page = Settings.ActiveTab;

            if (!Settings.Expanded)
                expander.Activate();

            // Show the window
            window.ShowAll();

            // Set small size request to prevent the status string to force the
            // main window to grow too large
            gxml.GetWidget("expander").SetSizeRequest(1, -1);

            // This is required to enable panning of the status string.
            // Otherwise the 'exapnder' steals all motion notify events.
            gxml.GetWidget("expander").AddEvents((int) Gdk.EventMask.ButtonMotionMask);

            // Register update delegates
            Connection.Instance.UpdateDelegates += Update;

            // Force first update and refresh every 500 ms
            PollMpd();
            GLib.Timeout.Add(500, PollMpd);

            // GConf notifications
            Settings.StatusFormatChanged += StatusFormatNotifyEventHandler;
            Settings.GuiRow1Changed += GuiRow1NotifyEventHandler;
            Settings.GuiRow2Changed += GuiRow2NotifyEventHandler;
        }

        void GuiRow1NotifyEventHandler(object o, GConf.NotifyEventArgs args)
        {
            SetGuiWidgets(hbox_row1, Settings.GuiRow1);
        }

        void GuiRow2NotifyEventHandler(object o, GConf.NotifyEventArgs args)
        {
            SetGuiWidgets(hbox_row2, Settings.GuiRow2);
        }

        /*
         * Updates
         *  - status tring
         *  - window title
         *  - status bar
         *
         * This is an update delegate just as all the others like the library
         * tab's or the cover tab's. One difference is that this is the only
         * one that is not optional - the status string and statusbar are not
         * hideable
         *
         */
        void Update()
        {
            // Status string and window title
            if (Connection.Instance.prevStatusInfo.Music.SongID !=
                    Connection.Instance.statusInfo.Music.SongID ||
                    Connection.Instance.prevStatusInfo.PlaylistVersion !=
                    Connection.Instance.statusInfo.PlaylistVersion)
            {
                UpdateStatusString();
                window.Title = Connection.Instance.musicInfo.Filename;
            }

            // Statusbar
            System.Text.StringBuilder sb = new System.Text.StringBuilder();
            switch(Connection.Instance.statusInfo.State)
            {
                case StateStatus.Play:
                    MusicStatus ms = Connection.Instance.statusInfo.Music;
                    sb.Append("Playing, ").Append(ms.BitRate).Append(" kbps, ");
                    sb.Append(ms.SampleRate).Append(" Hz");
                    break;
                case StateStatus.Pause:
                    sb.Append("Paused");
                    break;
                case StateStatus.Stop:
                    sb.Append("Stopped");
                    break;
            }
            statusbar.Push(1, sb.ToString());

            // Updating database?
            if (Connection.Instance.statusInfo.UpdatingDB)
            {
                statusbar.Push(1, "Updating database...");
            }
        }

        /*
         * Force an update of the parsed status string when we get a gconf
         * change notification.
         */
        void StatusFormatNotifyEventHandler(object o, GConf.NotifyEventArgs args)
        {
            UpdateStatusString();
        }

        /*
         * Add a tab of the specified type.
         * In principle, there is nothing which prevents several tabs of the
         * same type.
         */
        public void AddTab(TabType type)
        {
            // Construct the right tab...
            Widget widget = null;
            string title = "";
            switch(type)
            {
                case TabType.Playlist:
                    widget = new Playlist();
                    title = "_Playlist";
                    break;
                case TabType.Library:
                    widget = new Library();
                    title = "_Library";
                    break;
                case TabType.Find:
                    widget = new Find();
                    title = "_Find";
                    break;
                case TabType.Cover:
                    widget = new Cover();
                    title = "_Cover";
                    break;
                case TabType.Playlists:
                    widget = new Playlists();
                    title = "Playli_sts";
                    break;
            }

            // ...and add it
            if (widget != null)
            {
                notebook.AppendPage(widget, new Gtk.Label(title));
                notebook.ShowAll();
            }

            SaveTabOrder();

            // Set Connection.Instance.client to null to force a proper first update
            Connection.Instance.client = null;
        }

        /*
         * Remove the tab of the specified type.
         * At least one tab must be visible all the time.
         */
        void RemoveTab(Type type)
        {
            //// Don't remove the last tab
            //if (notebook.NPages <= 1)
            //    return;
            // Find a child of the same type
            foreach (Widget child in notebook.Children)
            {
                if (child.GetType() == type)
                {
                    // Unregister UpdateDelegate
                    if (child is IUpdatable)
                        (child as IUpdatable).UnregisterUpdateDelegate();
                    // Remove it
                    notebook.RemovePage(notebook.PageNum(child));
                }
            }
            SaveTabOrder();
        }

        /*
         * Parse the GConf settings format for one of the gui rows and
         * construct the found widgets.
         * All the widgets in the same row are removed and readded regardless
         * of its previous state.
         */
        void SetGuiWidgets(HBox hbox, Array settings)
        {
            // Remove all old widgets
            foreach (Widget w in hbox.Children)
            {
                if (w is IUpdatable)
                    (w as IUpdatable).UnregisterUpdateDelegate();
            }
            foreach (Widget w in hbox.Children)
            {
                hbox.Remove(w);
            }

            // Append the new ones
            foreach (string name in (string[]) settings)
            {
                Factory.AddGuiWidget(name, hbox);
            }

            hbox.ShowAll();
        }

        /*
         * This is the timeout method executed at short intervalls.
         * Here we try to connecto to mpd if not connected.
         * If a connection is established, all the status information from mpd
         * is updated. Often it's the difference between two consecutive
         * updates that dictates if something needs updating.
         *
         * Finally, here is where we execute all the currently registered
         * update delegates, thus making sure that all the diffrent parts of
         * the program are updated correctly.
         */
        bool PollMpd()
        {
            if (Connection.Instance.client == null)
            {
                if (!Connection.Instance.Connect())
                {
                    statusbar.Push(1, "Not connected");
                    Console.WriteLine("Connection failed");
                    return true;
                }

                if (!Connection.Instance.Refresh())
                {
                    Console.WriteLine("Refresh failed");
                    return true;
                }

                // We are now connected. Make sure that the 'old' status values
                // are such that a first proper update is performed.
                Connection.Instance.prevStatusInfo.Music.SongID = -1;
                Connection.Instance.prevStatusInfo.Music.Song = -1;
                Connection.Instance.prevStatusInfo.PlaylistVersion = 0;
                Connection.Instance.prevStatsInfo.DBUpdateDate = DateTime.MinValue;
                statusbar.Push(1, "Connected");
            }
            else
            {
                if (!Connection.Instance.Refresh())
                {
                    return true;
                }
            }

            // Run all registred update delegates
            Connection.Instance.RunUpdateDelegates();

            return true;
        }

        /*
         * Parse the status format string and show the resulting string (with
         * optional pango markup).
         */
        void UpdateStatusString()
        {
            Formatter.MusicInfoFormat = Settings.StatusFormat;
            string text = Formatter.MusicInfo(Connection.Instance.musicInfo);
            text = text.Replace(System.Environment.NewLine, "");
            text = text.Replace("\\n", System.Environment.NewLine);
            text = text.Replace("\\t", "\t");
            expander_label.Text = text;
            expander_label.UseMarkup = true;
        }

        /*
         * Main gui callbacks
         */
        [GLib.ConnectBefore]
        void OnExpanderButtonRelease(object o, ButtonReleaseEventArgs args)
        {
            if (args.Event.Button == 3)
            {
                //// TODO: Don't use the glade file (for now anyway). The only event
                //// that seem to be connectable in the glade gui builder is the
                //// 'Activated' event but here the 'Toggled' event seems to be
                //// required.
                ////Glade.XML gxml =
                ////    Glade.XML.FromAssembly("gui.glade", "move_tab_menu", null);
                ////gxml.Autoconnect(this);
                Menu tabConfigMenu = new Menu();
                ImageMenuItem moveLeft = new ImageMenuItem("Move left", null);
                ImageMenuItem moveRight = new ImageMenuItem("Move right", null);
                CheckMenuItem showPlaylist = new CheckMenuItem("Playlist");
                CheckMenuItem showLibrary = new CheckMenuItem("Library");
                CheckMenuItem showFind = new CheckMenuItem("Find");
                CheckMenuItem showCover = new CheckMenuItem("Cover");
                CheckMenuItem showPlaylists = new CheckMenuItem("Playlists");
                ImageMenuItem showPreferences = new ImageMenuItem(Gtk.Stock.Preferences, null);

                moveLeft.Image = new Image(Stock.GoBack, IconSize.Menu);
                moveRight.Image = new Image(Stock.GoForward, IconSize.Menu);

                tabConfigMenu.Append(moveLeft);
                tabConfigMenu.Append(moveRight);
                tabConfigMenu.Append(new SeparatorMenuItem());
                tabConfigMenu.Append(showPlaylist);
                tabConfigMenu.Append(showLibrary);
                tabConfigMenu.Append(showFind);
                tabConfigMenu.Append(showCover);
                tabConfigMenu.Append(showPlaylists);
                tabConfigMenu.Append(new SeparatorMenuItem());
                tabConfigMenu.Append(showPreferences);

                string[] tabs = Settings.Tabs as string[];
                showPlaylist.Active = System.Array.IndexOf(tabs, "Playlist") > -1;
                showLibrary.Active = System.Array.IndexOf(tabs, "Library") > -1;
                showFind.Active = System.Array.IndexOf(tabs, "Find") > -1;
                showCover.Active = System.Array.IndexOf(tabs, "Cover") > -1;
                showPlaylists.Active = System.Array.IndexOf(tabs, "Playlists") > -1;

                moveLeft.Activated += new EventHandler(OnMoveLeftActivate);
                moveRight.Activated += new EventHandler(OnMoveRightActivate);
                showPlaylist.Toggled += new EventHandler(OnShowPlaylistToggled);
                showLibrary.Toggled += new EventHandler(OnShowLibraryToggled);
                showFind.Toggled += new EventHandler(OnShowFindToggled);
                showCover.Toggled += new EventHandler(OnShowCoverToggled);
                showPlaylists.Toggled += new EventHandler(OnShowPlaylistsToggled);
                showPreferences.Activated += new EventHandler(OnShowOptions);

                tabConfigMenu.ShowAll();
                tabConfigMenu.Popup();
            }
            else if (args.Event.Button == 2)
            {
                Gtk.Grab.Remove(expander_eventbox);
            }
        }

        /*
         * Using mouse button 2 saves the click position and adjustment
         * position required to pan a long status title.
         */
        [GLib.ConnectBefore]
        void OnExpanderButtonPress(object o, ButtonPressEventArgs args)
        {
            if (args.Event.Button == 2)
            {
                PanBeginX = args.Event.XRoot;
                PanBeginHadj = expander_vp.Hadjustment.Value;
                Gtk.Grab.Add(expander_eventbox);
            }
        }

        /*
         * This is executed before the expander has actually changed state.
         * We must hide or show the contents before the state change.
         * When going from exapnded to non-expanded mode, we must save the
         * original window size. If not the correct expanded size can't be
         * restored at startup in case we close the program in non-expanded
         * mode.
         */
        [GLib.ConnectBefore]
        void OnExpanderActivated(object o, EventArgs args)
        {
            if (expander.Expanded)
            {
                int width, height;
                window.GetSize(out width, out height);
                Settings.WinWidth = width;
                Settings.WinHeight = height;
                expander.Child.HideAll();
            }
            else
            {
                expander.Child.ShowAll();
            }
        }

        /*
         * After the expanded status change, we must set the correct size.
         */
        void OnAfterExpanderActivated(object o, EventArgs args)
        {
            int width, height;
            window.GetSize(out width, out height);
            if (!expander.Expanded)
            {
                window.Resize(width, 1);
            }
            else
            {
                window.Resize(width, Settings.WinHeight);
            }
        }

        /*
         * Scroll the status string depending on how far we have dragged/panned
         * from the start position.
         */
        [GLib.ConnectBefore]
        void OnExpanderEventboxMotionNotify(object o, MotionNotifyEventArgs args)
        {
            double dx = args.Event.XRoot - PanBeginX;
            Adjustment hadj = expander_vp.Hadjustment;
            double new_hadj = System.Math.Min(PanBeginHadj - dx, hadj.Upper - hadj.PageSize);
            hadj.Value = new_hadj;
        }

        /*
         * When quitting the program, some configuration variables must be
         * saved.
         */
        void OnDeleteEvent(object o, DeleteEventArgs args)
        {
            // Save window position
            int pos_x, pos_y;
            window.GetPosition(out pos_x, out pos_y);
            Settings.WinPosX = pos_x;
            Settings.WinPosY = pos_y;

            // Save window size
            if (expander.Expanded)
            {
                int width, height;
                window.GetSize(out width, out height);
                Settings.WinWidth = width;
                Settings.WinHeight = height;
            }

            // Save expanded state
            Settings.Expanded = expander.Expanded;

            // Save current tab
            Settings.ActiveTab = notebook.Page;

            Application.Quit();
            args.RetVal = true;
        }

        /*
         * Capture global shortcut keys.
         */
        void OnKeyPress(object o, KeyPressEventArgs args)
        {
            Gdk.EventKey e = args.Event;

            // Quit - Escape
            if (e.Key == Gdk.Key.Escape)
            {
                Application.Quit();
            }
            // Settings - Ctrl-O
            else if (e.Key == Gdk.Key.o && (e.State & Gdk.ModifierType.ControlMask)
                    == Gdk.ModifierType.ControlMask)
            {
                new SettingsEditor();
            }
        }

        /*
         * Move tabs menu callbacks 
         */
        void OnMoveLeftActivate(object o, EventArgs args)
        {
            int newPos = System.Math.Max(0, notebook.Page -1);
            notebook.ReorderChild(notebook.CurrentPageWidget, newPos);
            SaveTabOrder();
        }

        void OnMoveRightActivate(object o, EventArgs args)
        {
            int newPos = System.Math.Min(notebook.Page + 1, notebook.NPages-1);
            notebook.ReorderChild(notebook.CurrentPageWidget, newPos);
            SaveTabOrder();
        }

        void SaveTabOrder()
        {
            ArrayList tabs = new ArrayList();
            for (int i=0; i < notebook.NPages; i++)
                tabs.Add(notebook.GetTabLabelText(notebook.GetNthPage(i)));
            if (tabs.Count > 0)
                Settings.Tabs = tabs.ToArray();
            else
            {
                string[] empty = new string[1];
                empty[0] = "";
                Settings.Tabs = empty;
            }
        }

        void OnShowPlaylistToggled(object o, EventArgs args)
        {
            if ((o as CheckMenuItem).Active)
                AddTab(TabType.Playlist);
            else
                RemoveTab(typeof(Playlist));
        }

        void OnShowLibraryToggled(object o, EventArgs args)
        {
            if ((o as CheckMenuItem).Active)
                AddTab(TabType.Library);
            else
                RemoveTab(typeof(Library));
        }

        void OnShowFindToggled(object o, EventArgs args)
        {
            if ((o as CheckMenuItem).Active)
                AddTab(TabType.Find);
            else
                RemoveTab(typeof(Find));
        }

        void OnShowCoverToggled(object o, EventArgs args)
        {
            if ((o as CheckMenuItem).Active)
                AddTab(TabType.Cover);
            else
                RemoveTab(typeof(Cover));
        }

        void OnShowPlaylistsToggled(object o, EventArgs args)
        {
            if ((o as CheckMenuItem).Active)
                AddTab(TabType.Playlists);
            else
                RemoveTab(typeof(Playlists));
        }

        void OnShowOptions(object o, EventArgs args)
        {
            new SettingsEditor();
        }
    }
}
